@@ -1,30 +1,9 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of Qt Creator.
-**
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 as published by the Free Software
-** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-****************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
 
 #include "basetextfind.h"
 
+#include <utils/algorithm.h>
 #include <utils/qtcassert.h>
 #include <utils/filesearch.h>
 
@@ -34,15 +13,17 @@
 #include <QTextBlock>
 #include <QTextCursor>
 
+using namespace Utils;
+
 namespace Core {
 
-static QRegularExpression regularExpression(const QString &txt, FindFlags flags)
+QRegularExpression BaseTextFind::regularExpression(const QString &txt, FindFlags flags)
 {
-    return QRegularExpression(
-                (flags & FindRegularExpression) ? txt
-                                                : QRegularExpression::escape(txt),
-                (flags & FindCaseSensitively) ? QRegularExpression::NoPatternOption
-                                              : QRegularExpression::CaseInsensitiveOption);
+    return QRegularExpression((flags & FindRegularExpression) ? txt
+                                                              : QRegularExpression::escape(txt),
+                              (flags & FindCaseSensitively)
+                                  ? QRegularExpression::NoPatternOption
+                                  : QRegularExpression::CaseInsensitiveOption);
 }
 
 struct BaseTextFindPrivate
@@ -53,10 +34,8 @@ struct BaseTextFindPrivate
     QPointer<QTextEdit> m_editor;
     QPointer<QPlainTextEdit> m_plaineditor;
     QPointer<QWidget> m_widget;
-    QTextCursor m_findScopeStart;
-    QTextCursor m_findScopeEnd;
-    int m_findScopeVerticalBlockSelectionFirstColumn;
-    int m_findScopeVerticalBlockSelectionLastColumn;
+    Utils::MultiTextCursor m_scope;
+    std::function<Utils::MultiTextCursor()> m_cursorProvider;
     int m_incrementalStartPos;
     bool m_incrementalWrappedState;
 };
@@ -64,8 +43,6 @@ struct BaseTextFindPrivate
 BaseTextFindPrivate::BaseTextFindPrivate(QTextEdit *editor)
     : m_editor(editor)
     , m_widget(editor)
-    , m_findScopeVerticalBlockSelectionFirstColumn(-1)
-    , m_findScopeVerticalBlockSelectionLastColumn(-1)
     , m_incrementalStartPos(-1)
     , m_incrementalWrappedState(false)
 {
@@ -74,8 +51,6 @@ BaseTextFindPrivate::BaseTextFindPrivate(QTextEdit *editor)
 BaseTextFindPrivate::BaseTextFindPrivate(QPlainTextEdit *editor)
     : m_plaineditor(editor)
     , m_widget(editor)
-    , m_findScopeVerticalBlockSelectionFirstColumn(-1)
-    , m_findScopeVerticalBlockSelectionLastColumn(-1)
     , m_incrementalStartPos(-1)
     , m_incrementalWrappedState(false)
 {
@@ -93,19 +68,14 @@ BaseTextFindPrivate::BaseTextFindPrivate(QPlainTextEdit *editor)
 */
 
 /*!
-    \fn void Core::BaseTextFind::findScopeChanged(const QTextCursor &start,
-                           const QTextCursor &end,
-                           int verticalBlockSelectionFirstColumn,
-                           int verticalBlockSelectionLastColumn)
+    \fn void Core::BaseTextFind::findScopeChanged(const Utils::MultiTextCursor &cursor)
 
     This signal is emitted when the search
-    scope changes to \a start, \a end,
-    \a verticalBlockSelectionFirstColumn, and
-    \a verticalBlockSelectionLastColumn.
+    scope changes to \a cursor.
 */
 
 /*!
-    \fn void Core::BaseTextFind::highlightAllRequested(const QString &txt, Core::FindFlags findFlags)
+    \fn void Core::BaseTextFind::highlightAllRequested(const QString &txt, Utils::FindFlags findFlags)
 
     This signal is emitted when the search results for \a txt using the given
     \a findFlags should be highlighted in the editor widget.
@@ -209,7 +179,7 @@ QString BaseTextFind::currentFindString() const
         cursor.movePosition(QTextCursor::StartOfWord);
         cursor.movePosition(QTextCursor::EndOfWord, QTextCursor::KeepAnchor);
         QString s = cursor.selectedText();
-        foreach (QChar c, s) {
+        for (const QChar c : s) {
             if (!c.isLetterOrNumber() && c != QLatin1Char('_')) {
                 s.clear();
                 break;
@@ -279,7 +249,7 @@ void BaseTextFind::replace(const QString &before, const QString &after, FindFlag
     setTextCursor(cursor);
 }
 
-// QTextCursor::insert moves all other QTextCursors that are the the insertion point forward.
+// QTextCursor::insert moves all other QTextCursors that are the insertion point forward.
 // We do not want that for the replace operation, because then e.g. the find scope would move when
 // replacing a match at the start.
 static void insertTextAfterSelection(const QString &text, QTextCursor &cursor)
@@ -324,6 +294,13 @@ QTextCursor BaseTextFind::replaceInternal(const QString &before, const QString &
     return cursor;
 }
 
+Utils::MultiTextCursor BaseTextFind::multiTextCursor() const
+{
+    if (d->m_cursorProvider)
+        return d->m_cursorProvider();
+    return Utils::MultiTextCursor({textCursor()});
+}
+
 /*!
     \reimp
 */
@@ -344,18 +321,20 @@ bool BaseTextFind::replaceStep(const QString &before, const QString &after, Find
 int BaseTextFind::replaceAll(const QString &before, const QString &after, FindFlags findFlags)
 {
     QTextCursor editCursor = textCursor();
-    if (!d->m_findScopeStart.isNull())
-        editCursor.setPosition(d->m_findScopeStart.position());
+    if (findFlags.testFlag(FindBackward))
+        editCursor.movePosition(QTextCursor::End);
     else
         editCursor.movePosition(QTextCursor::Start);
+    editCursor.movePosition(QTextCursor::Start);
     editCursor.beginEditBlock();
     int count = 0;
     bool usesRegExp = (findFlags & FindRegularExpression);
     bool preserveCase = (findFlags & FindPreserveCase);
     QRegularExpression regexp = regularExpression(before, findFlags);
-    QTextCursor found = findOne(regexp, editCursor, textDocumentFlagsForFindFlags(findFlags));
+    QTextCursor found = findOne(regexp, editCursor,
+                                Utils::textDocumentFlagsForFindFlags(findFlags));
     bool first = true;
-    while (!found.isNull() && inScope(found.selectionStart(), found.selectionEnd())) {
+    while (!found.isNull()) {
         if (found == editCursor && !first) {
             if (editCursor.atEnd())
                 break;
@@ -366,7 +345,7 @@ int BaseTextFind::replaceAll(const QString &before, const QString &after, FindFl
             newPosCursor.movePosition(findFlags & FindBackward ?
                                           QTextCursor::PreviousCharacter :
                                           QTextCursor::NextCharacter);
-            found = findOne(regexp, newPosCursor, textDocumentFlagsForFindFlags(findFlags));
+            found = findOne(regexp, newPosCursor, Utils::textDocumentFlagsForFindFlags(findFlags));
             continue;
         }
         if (first)
@@ -384,103 +363,80 @@ int BaseTextFind::replaceAll(const QString &before, const QString &after, FindFl
         else
             realAfter = after;
         insertTextAfterSelection(realAfter, editCursor);
-        found = findOne(regexp, editCursor, textDocumentFlagsForFindFlags(findFlags));
+        found = findOne(regexp, editCursor, Utils::textDocumentFlagsForFindFlags(findFlags));
     }
     editCursor.endEditBlock();
     return count;
 }
 
-bool BaseTextFind::find(const QString &txt, FindFlags findFlags,
-    QTextCursor start, bool *wrapped)
+bool BaseTextFind::find(const QString &txt, FindFlags findFlags, QTextCursor start, bool *wrapped)
 {
     if (txt.isEmpty()) {
         setTextCursor(start);
         return true;
     }
     QRegularExpression regexp = regularExpression(txt, findFlags);
-    QTextCursor found = findOne(regexp, start, textDocumentFlagsForFindFlags(findFlags));
+    QTextCursor found = findOne(regexp, start, Utils::textDocumentFlagsForFindFlags(findFlags));
     if (wrapped)
         *wrapped = false;
 
-    if (!d->m_findScopeStart.isNull()) {
-
-        // scoped
-        if (found.isNull() || !inScope(found.selectionStart(), found.selectionEnd())) {
-            if ((findFlags & FindBackward) == 0)
-                start.setPosition(d->m_findScopeStart.position());
-            else
-                start.setPosition(d->m_findScopeEnd.position());
-            found = findOne(regexp, start, textDocumentFlagsForFindFlags(findFlags));
-            if (found.isNull() || !inScope(found.selectionStart(), found.selectionEnd()))
-                return false;
-            if (wrapped)
-                *wrapped = true;
-        }
-    } else {
-
-        // entire document
-        if (found.isNull()) {
-            if ((findFlags & FindBackward) == 0)
-                start.movePosition(QTextCursor::Start);
-            else
-                start.movePosition(QTextCursor::End);
-            found = findOne(regexp, start, textDocumentFlagsForFindFlags(findFlags));
-            if (found.isNull())
-                return false;
-            if (wrapped)
-                *wrapped = true;
-        }
+    if (found.isNull()) {
+        if ((findFlags & FindBackward) == 0)
+            start.movePosition(QTextCursor::Start);
+        else
+            start.movePosition(QTextCursor::End);
+        found = findOne(regexp, start, Utils::textDocumentFlagsForFindFlags(findFlags));
+        if (found.isNull())
+            return false;
+        if (wrapped)
+            *wrapped = true;
     }
-    if (!found.isNull())
-        setTextCursor(found);
+    setTextCursor(found);
     return true;
 }
 
-
-// helper function. Works just like QTextDocument::find() but supports vertical block selection
 QTextCursor BaseTextFind::findOne(const QRegularExpression &expr,
-                                  const QTextCursor &from, QTextDocument::FindFlags options) const
+                                  QTextCursor from,
+                                  QTextDocument::FindFlags options) const
 {
-    QTextCursor candidate = document()->find(expr, from, options);
-    if (candidate.isNull())
-        return candidate;
-
-    if (d->m_findScopeVerticalBlockSelectionFirstColumn < 0)
-        return candidate;
-    forever {
-        if (!inScope(candidate.selectionStart(), candidate.selectionEnd()))
-            return candidate;
-        bool inVerticalFindScope = false;
-        // This code relies on the fact, that we have to keep TextEditorWidget subclass
-        // inside d->m_plaineditor which is of QPlainTextEdit class. So we can't
-        // transform it into a typed version now, as it relies on a dynamic match.
-        QMetaObject::invokeMethod(d->m_plaineditor, "inFindScope", Qt::DirectConnection,
-                                  Q_RETURN_ARG(bool, inVerticalFindScope),
-                                  Q_ARG(QTextCursor, candidate));
-        if (inVerticalFindScope)
-            return candidate;
-
-        QTextCursor newCandidate = document()->find(expr, candidate, options);
-        if (newCandidate == candidate) {
-            // When searching for regular expressions that match "zero length" strings (like ^ or \b)
-            // we need to move away from the match before searching for the next one.
-            candidate.movePosition(options & QTextDocument::FindBackward
-                                   ? QTextCursor::PreviousCharacter
-                                   : QTextCursor::NextCharacter);
-            candidate = document()->find(expr, candidate, options);
+    QTextCursor found = document()->find(expr, from, options);
+    while (!found.isNull() && !inScope(found)) {
+        if (!found.hasSelection()) {
+            if (found.movePosition(options & QTextDocument::FindBackward
+                                       ? QTextCursor::PreviousCharacter
+                                       : QTextCursor::NextCharacter)) {
+                from = found;
+            } else {
+                return {};
+            }
         } else {
-            candidate = newCandidate;
+            from.setPosition(options & QTextDocument::FindBackward ? found.selectionStart()
+                                                                   : found.selectionEnd());
         }
+        found = document()->find(expr, from, options);
     }
-    return candidate;
+
+    return found;
 }
 
-bool BaseTextFind::inScope(int startPosition, int endPosition) const
+bool BaseTextFind::inScope(const QTextCursor &candidate) const
 {
-    if (d->m_findScopeStart.isNull())
+    if (candidate.isNull())
+        return false;
+    if (d->m_scope.isNull())
         return true;
-    return (d->m_findScopeStart.position() <= startPosition
-            && d->m_findScopeEnd.position() >= endPosition);
+    return inScope(candidate.selectionStart(), candidate.selectionEnd());
+}
+
+bool BaseTextFind::inScope(int candidateStart, int candidateEnd) const
+{
+    if (d->m_scope.isNull())
+        return true;
+    if (candidateStart > candidateEnd)
+        std::swap(candidateStart, candidateEnd);
+    return Utils::anyOf(d->m_scope, [&](const QTextCursor &scope) {
+        return candidateStart >= scope.selectionStart() && candidateEnd <= scope.selectionEnd();
+    });
 }
 
 /*!
@@ -488,30 +444,23 @@ bool BaseTextFind::inScope(int startPosition, int endPosition) const
 */
 void BaseTextFind::defineFindScope()
 {
-    QTextCursor cursor = textCursor();
-    if (cursor.hasSelection() && cursor.block() != cursor.document()->findBlock(cursor.anchor())) {
-        d->m_findScopeStart = cursor;
-        d->m_findScopeStart.setPosition(qMax(0, cursor.selectionStart()));
-        d->m_findScopeEnd = cursor;
-        d->m_findScopeEnd.setPosition(cursor.selectionEnd());
-        d->m_findScopeVerticalBlockSelectionFirstColumn = -1;
-        d->m_findScopeVerticalBlockSelectionLastColumn = -1;
-
-        if (d->m_plaineditor && d->m_plaineditor->metaObject()->indexOfProperty("verticalBlockSelectionFirstColumn") >= 0) {
-            d->m_findScopeVerticalBlockSelectionFirstColumn
-                    = d->m_plaineditor->property("verticalBlockSelectionFirstColumn").toInt();
-            d->m_findScopeVerticalBlockSelectionLastColumn
-                    = d->m_plaineditor->property("verticalBlockSelectionLastColumn").toInt();
+    Utils::MultiTextCursor multiCursor = multiTextCursor();
+    bool foundSelection = false;
+    for (const QTextCursor &c : multiCursor) {
+        if (c.hasSelection()) {
+            if (foundSelection || c.block() != c.document()->findBlock(c.anchor())) {
+                const QList<QTextCursor> sortedCursors = Utils::sorted(multiCursor.cursors());
+                d->m_scope = Utils::MultiTextCursor(sortedCursors);
+                QTextCursor cursor = textCursor();
+                cursor.clearSelection();
+                setTextCursor(cursor);
+                emit findScopeChanged(d->m_scope);
+                return;
+            }
+            foundSelection = true;
         }
-
-        emit findScopeChanged(d->m_findScopeStart, d->m_findScopeEnd,
-                              d->m_findScopeVerticalBlockSelectionFirstColumn,
-                              d->m_findScopeVerticalBlockSelectionLastColumn);
-        cursor.setPosition(d->m_findScopeStart.position());
-        setTextCursor(cursor);
-    } else {
-        clearFindScope();
     }
+    clearFindScope();
 }
 
 /*!
@@ -519,13 +468,8 @@ void BaseTextFind::defineFindScope()
 */
 void BaseTextFind::clearFindScope()
 {
-    d->m_findScopeStart = QTextCursor();
-    d->m_findScopeEnd = QTextCursor();
-    d->m_findScopeVerticalBlockSelectionFirstColumn = -1;
-    d->m_findScopeVerticalBlockSelectionLastColumn = -1;
-    emit findScopeChanged(d->m_findScopeStart, d->m_findScopeEnd,
-                          d->m_findScopeVerticalBlockSelectionFirstColumn,
-                          d->m_findScopeVerticalBlockSelectionLastColumn);
+    d->m_scope = Utils::MultiTextCursor();
+    emit findScopeChanged(d->m_scope);
 }
 
 /*!
@@ -537,4 +481,9 @@ void BaseTextFind::highlightAll(const QString &txt, FindFlags findFlags)
     emit highlightAllRequested(txt, findFlags);
 }
 
+void BaseTextFind::setMultiTextCursorProvider(const CursorProvider &provider)
+{
+    d->m_cursorProvider = provider;
+}
+
 } // namespace Core
