@@ -28,8 +28,11 @@
 #include <coreplugin/idocument.h>
 #include <coreplugin/icore.h>
 
+#include <projectexplorer/runcontrol.h>
+
+#include <utils/algorithm.h>
 #include <utils/environment.h>
-#include <utils/process.h>
+#include <utils/qtcprocess.h>
 #include <utils/processinterface.h>
 #include <utils/qtcassert.h>
 
@@ -179,8 +182,8 @@ void LldbEngine::setupEngine()
 
     showMessage("STARTING LLDB: " + lldbCmd.toUserOutput());
     Environment environment = runParameters().debugger.environment;
-    environment.appendOrSet("QT_CREATOR_LLDB_PROCESS", "1");
-    environment.appendOrSet("PYTHONUNBUFFERED", "1");  // avoid flushing problem on macOS
+    environment.set("QT_CREATOR_LLDB_PROCESS", "1");
+    environment.set("PYTHONUNBUFFERED", "1"); // avoid flushing problem on macOS
     DebuggerItem::addAndroidLldbPythonEnv(lldbCmd, environment);
 
     if (lldbCmd.osType() == OsTypeLinux) {
@@ -198,6 +201,11 @@ void LldbEngine::setupEngine()
             environment.appendOrSet("PYTHONPATH", "/usr/lib/llvm-14/lib/python3.10/dist-packages");
     }
 
+    if (runParameters().runAsRoot) {
+        ProjectExplorer::RunControl::provideAskPassEntry(environment);
+        m_lldbProc.setRunAsRoot(true);
+    }
+
     m_lldbProc.setEnvironment(environment);
 
     if (runParameters().debugger.workingDirectory.isDir())
@@ -213,7 +221,8 @@ void LldbEngine::setupEngine()
 
 void LldbEngine::handleLldbStarted()
 {
-    m_lldbProc.waitForReadyRead(1000);
+    using namespace std::chrono_literals;
+    m_lldbProc.waitForReadyRead(1s);
 
     showStatusMessage(Tr::tr("Setting up inferior..."));
 
@@ -229,7 +238,7 @@ void LldbEngine::handleLldbStarted()
         executeCommand(commands);
 
     const FilePath path = settings().extraDumperFile();
-    if (!path.isEmpty() && path.isReadableFile()) {
+    if (path.isReadableFile()) {
         DebuggerCommand cmd("addDumperModule");
         cmd.arg("path", path.path());
         runCommand(cmd);
@@ -266,7 +275,13 @@ void LldbEngine::handleLldbStarted()
     cmd2.arg("startmode", rp.startMode);
     cmd2.arg("nativemixed", isNativeMixedActive());
     cmd2.arg("workingdirectory", rp.inferior.workingDirectory.path());
-    cmd2.arg("environment", rp.inferior.environment.toStringList());
+    QStringList environment = rp.inferior.environment.toStringList();
+    // Prevent lldb from automatically setting OS_ACTIVITY_DT_MODE to mirror
+    // NSLog to stderr, as that will also mirror os_log, which we pick up in
+    // AppleUnifiedLogger::preventsStderrLogging(), and end up disabling Qt's
+    // default stderr logger. We prefer Qt's own stderr logging if we can.
+    environment << "IDE_DISABLED_OS_ACTIVITY_DT_MODE=1";
+    cmd2.arg("environment", environment);
     cmd2.arg("processargs", toHex(ProcessArgs::splitArgs(rp.inferior.command.arguments(),
                                                          HostOsInfo::hostOs()).join(QChar(0))));
     cmd2.arg("platform", rp.platform);
@@ -283,21 +298,23 @@ void LldbEngine::handleLldbStarted()
         cmd2.arg("attachpid", attachedPID);
     } else {
         cmd2.arg("startmode", rp.startMode);
-        // it is better not to check the start mode on the python sid (as we would have to duplicate the
-        // enum values), and thus we assume that if the rp.attachPID is valid we really have to attach
-        QTC_CHECK(rp.attachPID.isValid() && (rp.startMode == AttachToRemoteProcess
-                                             || rp.startMode == AttachToLocalProcess
-                                             || rp.startMode == AttachToRemoteServer));
-        cmd2.arg("attachpid", rp.attachPID.pid());
-        cmd2.arg("sysroot", rp.deviceSymbolsRoot.isEmpty() ? rp.sysRoot.toString()
-                                                           : rp.deviceSymbolsRoot);
-        cmd2.arg("remotechannel", ((rp.startMode == AttachToRemoteProcess
-                                   || rp.startMode == AttachToRemoteServer)
-                                  ? rp.remoteChannel : QString()));
-        QTC_CHECK(!rp.continueAfterAttach || (rp.startMode == AttachToRemoteProcess
-                                              || rp.startMode == AttachToLocalProcess
-                                              || rp.startMode == AttachToRemoteServer));
-        m_continueAtNextSpontaneousStop = false;
+        if (rp.startMode != StartInternal) {
+            // it is better not to check the start mode on the python sid (as we would have to duplicate the
+            // enum values), and thus we assume that if the rp.attachPID is valid we really have to attach
+            QTC_CHECK(rp.attachPID.isValid() && (rp.startMode == AttachToRemoteProcess
+                                                 || rp.startMode == AttachToLocalProcess
+                                                 || rp.startMode == AttachToRemoteServer));
+            cmd2.arg("attachpid", rp.attachPID.pid());
+            cmd2.arg("sysroot", rp.deviceSymbolsRoot.isEmpty() ? rp.sysRoot.toString()
+                                                               : rp.deviceSymbolsRoot);
+            cmd2.arg("remotechannel", ((rp.startMode == AttachToRemoteProcess
+                                       || rp.startMode == AttachToRemoteServer)
+                                      ? rp.remoteChannel : QString()));
+            QTC_CHECK(!rp.continueAfterAttach || (rp.startMode == AttachToRemoteProcess
+                                                  || rp.startMode == AttachToLocalProcess
+                                                  || rp.startMode == AttachToRemoteServer));
+            m_continueAtNextSpontaneousStop = false;
+        }
     }
 
     cmd2.callback = [this](const DebuggerResponse &response) {
@@ -312,6 +329,16 @@ void LldbEngine::handleLldbStarted()
                 runEngine();
             };
             runCommand(cmd3);
+
+            // Execute post attach commands
+            QStringList commands = settings().gdbPostAttachCommands().split('\n');
+            commands = Utils::filtered(commands, [](const QString line) {
+                const QString trimmed = line.trimmed();
+                return !trimmed.isEmpty() && !trimmed.startsWith('#');
+            });
+            for (const QString &cmd : commands) {
+                executeDebuggerCommand(cmd);
+            }
         } else {
             notifyEngineSetupFailed();
         }
@@ -319,10 +346,6 @@ void LldbEngine::handleLldbStarted()
 
     cmd2.flags = Silent;
     runCommand(cmd2);
-
-    DebuggerCommand cmd0("setFallbackQtVersion");
-    cmd0.arg("version", rp.fallbackQtVersion);
-    runCommand(cmd0);
 }
 
 void LldbEngine::runEngine()
@@ -380,9 +403,10 @@ void LldbEngine::handleResponse(const QString &response)
         const QString name = item.name();
         if (name == "result") {
             QString msg = item["status"].data();
-            if (!msg.isEmpty())
+            if (!msg.isEmpty()) {
                 msg[0] = msg.at(0).toUpper();
-            showStatusMessage(msg);
+                showStatusMessage(msg);
+            }
 
             int token = item["token"].toInt();
             showMessage(QString("%1^").arg(token), LogOutput);
@@ -740,6 +764,8 @@ void LldbEngine::doUpdateLocals(const UpdateParameters &params)
     cmd.arg("partialvar", params.partialVariable);
     cmd.arg("qobjectnames", s.showQObjectNames());
     cmd.arg("timestamps", s.logTimeStamps());
+    cmd.arg("qtversion", runParameters().qtVersion);
+    cmd.arg("qtnamespace", runParameters().qtNamespace);
 
     StackFrame frame = stackHandler()->currentFrame();
     cmd.arg("context", frame.context);
@@ -832,18 +858,23 @@ void LldbEngine::readLldbStandardError()
 
 void LldbEngine::readLldbStandardOutput()
 {
-    QByteArray outba = m_lldbProc.readAllRawStandardOutput();
-    outba.replace("\r\n", "\n");
-    QString out = QString::fromUtf8(outba);
-    showMessage(out, LogOutput);
+    const QByteArray out = m_lldbProc.readAllRawStandardOutput();
+    showMessage(QString::fromUtf8(out), LogOutput);
     m_inbuffer.append(out);
     while (true) {
-        int pos = m_inbuffer.indexOf("@\n");
-        if (pos == -1)
-            break;
-        QString response = m_inbuffer.left(pos).trimmed();
-        m_inbuffer = m_inbuffer.mid(pos + 2);
-        emit outputReady(response);
+        if (int pos = m_inbuffer.indexOf("@\n"); pos >= 0) {
+            const QByteArray response = m_inbuffer.left(pos).trimmed();
+            m_inbuffer = m_inbuffer.mid(pos + 2);
+            emit outputReady(QString::fromUtf8(response));
+            continue;
+        }
+        if (int pos = m_inbuffer.indexOf("@\r\n"); pos >= 0) {
+            const QByteArray response = m_inbuffer.left(pos).trimmed();
+            m_inbuffer = m_inbuffer.mid(pos + 3);
+            emit outputReady(QString::fromUtf8(response));
+            continue;
+        }
+        break;
     }
 }
 
@@ -1031,7 +1062,7 @@ void LldbEngine::fetchFullBacktrace()
 {
     DebuggerCommand cmd("fetchFullBacktrace");
     cmd.callback = [](const DebuggerResponse &response) {
-        Internal::openTextEditor("Backtrace $", fromHex(response.data["fulltrace"].data()));
+        Internal::openTextEditor("Backtrace$", fromHex(response.data["fulltrace"].data()));
     };
     runCommand(cmd);
 }
