@@ -1,52 +1,34 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of Qt Creator.
-**
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 as published by the Free Software
-** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-****************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
 
 #include "devicemanager.h"
 
 #include "idevicefactory.h"
+#include "../projectexplorertr.h"
+#include "../projectexplorerconstants.h"
 
 #include <coreplugin/icore.h>
-#include <coreplugin/messagemanager.h>
 
-#include <projectexplorer/projectexplorerconstants.h>
 #include <utils/algorithm.h>
-#include <utils/fileutils.h>
+#include <utils/devicefileaccess.h>
+#include <utils/environment.h>
+#include <utils/fsengine/fsengine.h>
 #include <utils/persistentsettings.h>
-#include <utils/portlist.h>
+#include <utils/qtcprocess.h>
 #include <utils/qtcassert.h>
 #include <utils/stringutils.h>
+#include <utils/terminalhooks.h>
 
-#include <QFileInfo>
 #include <QHash>
-#include <QList>
-#include <QString>
+#include <QMutex>
+#include <QMutexLocker>
 #include <QVariantList>
 
 #include <limits>
 #include <memory>
 
+using namespace Utils;
+
 namespace ProjectExplorer {
 namespace Internal {
 
@@ -59,7 +41,7 @@ class DeviceManagerPrivate
 public:
     DeviceManagerPrivate() = default;
 
-    int indexForId(Core::Id id) const
+    int indexForId(Id id) const
     {
         for (int i = 0; i < devices.count(); ++i) {
             if (devices.at(i)->id() == id)
@@ -68,11 +50,18 @@ public:
         return -1;
     }
 
-    static DeviceManager *clonedInstance;
-    QList<IDevice::Ptr> devices;
-    QHash<Core::Id, Core::Id> defaultDevices;
+    QList<IDevice::Ptr> deviceList() const
+    {
+        QMutexLocker locker(&mutex);
+        return devices;
+    }
 
-    Utils::PersistentSettingsWriter *writer = nullptr;
+    static DeviceManager *clonedInstance;
+
+    mutable QMutex mutex;
+    QList<IDevice::Ptr> devices;
+    QHash<Id, Id> defaultDevices;
+    PersistentSettingsWriter *writer = nullptr;
 };
 DeviceManager *DeviceManagerPrivate::clonedInstance = nullptr;
 
@@ -94,7 +83,19 @@ int DeviceManager::deviceCount() const
 
 void DeviceManager::replaceInstance()
 {
-    copy(DeviceManagerPrivate::clonedInstance, instance(), false);
+    const QList<Id> newIds =
+        Utils::transform(DeviceManagerPrivate::clonedInstance->d->devices, &IDevice::id);
+
+    for (const IDevice::Ptr &dev : std::as_const(m_instance->d->devices)) {
+        if (!newIds.contains(dev->id()))
+            dev->aboutToBeRemoved();
+    }
+
+    {
+        QMutexLocker locker(&instance()->d->mutex);
+        copy(DeviceManagerPrivate::clonedInstance, instance(), false);
+    }
+
     emit instance()->deviceListReplaced();
     emit instance()->updated();
 }
@@ -114,10 +115,15 @@ DeviceManager *DeviceManager::cloneInstance()
     return DeviceManagerPrivate::clonedInstance;
 }
 
+DeviceManager *DeviceManager::clonedInstance()
+{
+    return DeviceManagerPrivate::clonedInstance;
+}
+
 void DeviceManager::copy(const DeviceManager *source, DeviceManager *target, bool deep)
 {
     if (deep) {
-        foreach (const IDevice::ConstPtr &device, source->d->devices)
+        for (const IDevice::Ptr &device : std::as_const(source->d->devices))
             target->d->devices << device->clone();
     } else {
         target->d->devices = source->d->devices;
@@ -129,9 +135,19 @@ void DeviceManager::save()
 {
     if (d->clonedInstance == this || !d->writer)
         return;
-    QVariantMap data;
-    data.insert(QLatin1String(DeviceManagerKey), toMap());
-    d->writer->save(data, Core::ICore::mainWindow());
+    Store data;
+    data.insert(DeviceManagerKey, variantFromStore(toMap()));
+    d->writer->save(data, Core::ICore::dialogParent());
+}
+
+static FilePath settingsFilePath(const QString &extension)
+{
+    return Core::ICore::userResourcePath(extension);
+}
+
+static FilePath systemSettingsFilePath(const QString &deviceFileRelativePath)
+{
+    return Core::ICore::installerResourcePath(deviceFileRelativePath);
 }
 
 void DeviceManager::load()
@@ -139,25 +155,23 @@ void DeviceManager::load()
     QTC_ASSERT(!d->writer, return);
 
     // Only create writer now: We do not want to save before the settings were read!
-    d->writer = new Utils::PersistentSettingsWriter(
-                settingsFilePath(QLatin1String("/devices.xml")),
-                QLatin1String("QtCreatorDevices"));
+    d->writer = new PersistentSettingsWriter(settingsFilePath("devices.xml"), "QtCreatorDevices");
 
-    Utils::PersistentSettingsReader reader;
+    PersistentSettingsReader reader;
     // read devices file from global settings path
-    QHash<Core::Id, Core::Id> defaultDevices;
+    QHash<Id, Id> defaultDevices;
     QList<IDevice::Ptr> sdkDevices;
-    if (reader.load(systemSettingsFilePath(QLatin1String("/devices.xml"))))
-        sdkDevices = fromMap(reader.restoreValues().value(DeviceManagerKey).toMap(), &defaultDevices);
+    if (reader.load(systemSettingsFilePath("devices.xml")))
+        sdkDevices = fromMap(storeFromVariant(reader.restoreValues().value(DeviceManagerKey)), &defaultDevices);
     // read devices file from user settings path
     QList<IDevice::Ptr> userDevices;
-    if (reader.load(settingsFilePath(QLatin1String("/devices.xml"))))
-        userDevices = fromMap(reader.restoreValues().value(DeviceManagerKey).toMap(), &defaultDevices);
+    if (reader.load(settingsFilePath("devices.xml")))
+        userDevices = fromMap(storeFromVariant(reader.restoreValues().value(DeviceManagerKey)), &defaultDevices);
     // Insert devices into the model. Prefer the higher device version when there are multiple
     // devices with the same id.
-    foreach (IDevice::Ptr device, userDevices) {
-        foreach (const IDevice::Ptr &sdkDevice, sdkDevices) {
-            if (device->id() == sdkDevice->id()) {
+    for (IDevice::ConstPtr device : std::as_const(userDevices)) {
+        for (const IDevice::Ptr &sdkDevice : std::as_const(sdkDevices)) {
+            if (device->id() == sdkDevice->id() || device->rootPath() == sdkDevice->rootPath()) {
                 if (device->version() < sdkDevice->version())
                     device = sdkDevice;
                 sdkDevices.removeOne(sdkDevice);
@@ -167,7 +181,7 @@ void DeviceManager::load()
         addDevice(device);
     }
     // Append the new SDK devices to the model.
-    foreach (const IDevice::Ptr &sdkDevice, sdkDevices)
+    for (const IDevice::Ptr &sdkDevice : std::as_const(sdkDevices))
         addDevice(sdkDevice);
 
     // Overwrite with the saved default devices.
@@ -180,19 +194,33 @@ void DeviceManager::load()
     emit devicesLoaded();
 }
 
-QList<IDevice::Ptr> DeviceManager::fromMap(const QVariantMap &map,
-                                           QHash<Core::Id, Core::Id> *defaultDevices)
+static const IDeviceFactory *restoreFactory(const Store &map)
+{
+    const Id deviceType = IDevice::typeFromMap(map);
+    IDeviceFactory *factory = Utils::findOrDefault(IDeviceFactory::allDeviceFactories(),
+        [&map, deviceType](IDeviceFactory *factory) {
+            return factory->canRestore(map) && factory->deviceType() == deviceType;
+        });
+
+    if (!factory)
+        qWarning("Warning: No factory found for device '%s' of type '%s'.",
+                 qPrintable(IDevice::idFromMap(map).toString()),
+                 qPrintable(IDevice::typeFromMap(map).toString()));
+    return factory;
+}
+
+QList<IDevice::Ptr> DeviceManager::fromMap(const Store &map, QHash<Id, Id> *defaultDevices)
 {
     QList<IDevice::Ptr> devices;
 
     if (defaultDevices) {
-        const QVariantMap defaultDevsMap = map.value(DefaultDevicesKey).toMap();
+        const Store defaultDevsMap = storeFromVariant(map.value(DefaultDevicesKey));
         for (auto it = defaultDevsMap.constBegin(); it != defaultDevsMap.constEnd(); ++it)
-            defaultDevices->insert(Core::Id::fromString(it.key()), Core::Id::fromSetting(it.value()));
+            defaultDevices->insert(Id::fromString(stringFromKey(it.key())), Id::fromSetting(it.value()));
     }
-    const QVariantList deviceList = map.value(QLatin1String(DeviceListKey)).toList();
-    foreach (const QVariant &v, deviceList) {
-        const QVariantMap map = v.toMap();
+    const QVariantList deviceList = map.value(DeviceListKey).toList();
+    for (const QVariant &v : deviceList) {
+        const Store map = storeFromVariant(v);
         const IDeviceFactory * const factory = restoreFactory(map);
         if (!factory)
             continue;
@@ -204,45 +232,34 @@ QList<IDevice::Ptr> DeviceManager::fromMap(const QVariantMap &map,
     return devices;
 }
 
-QVariantMap DeviceManager::toMap() const
+Store DeviceManager::toMap() const
 {
-    QVariantMap map;
-    QVariantMap defaultDeviceMap;
-    using TypeIdHash = QHash<Core::Id, Core::Id>;
-    for (TypeIdHash::ConstIterator it = d->defaultDevices.constBegin();
-             it != d->defaultDevices.constEnd(); ++it) {
-        defaultDeviceMap.insert(it.key().toString(), it.value().toSetting());
-    }
-    map.insert(QLatin1String(DefaultDevicesKey), defaultDeviceMap);
+    Store map;
+    Store defaultDeviceMap;
+    for (auto it = d->defaultDevices.constBegin(); it != d->defaultDevices.constEnd(); ++it)
+        defaultDeviceMap.insert(keyFromString(it.key().toString()), it.value().toSetting());
+
+    map.insert(DefaultDevicesKey, variantFromStore(defaultDeviceMap));
     QVariantList deviceList;
-    foreach (const IDevice::ConstPtr &device, d->devices)
-        deviceList << device->toMap();
-    map.insert(QLatin1String(DeviceListKey), deviceList);
+    for (const IDevice::Ptr &device : std::as_const(d->devices))
+        deviceList << variantFromStore(device->toMap());
+    map.insert(DeviceListKey, deviceList);
     return map;
 }
 
-Utils::FilePath DeviceManager::settingsFilePath(const QString &extension)
-{
-    return Utils::FilePath::fromString(Core::ICore::userResourcePath() + extension);
-}
-
-Utils::FilePath DeviceManager::systemSettingsFilePath(const QString &deviceFileRelativePath)
-{
-    return Utils::FilePath::fromString(Core::ICore::installerResourcePath()
-                                       + deviceFileRelativePath);
-}
-
 void DeviceManager::addDevice(const IDevice::ConstPtr &_device)
 {
     const IDevice::Ptr device = _device->clone();
 
     QStringList names;
-    foreach (const IDevice::ConstPtr &tmp, d->devices) {
+    for (const IDevice::Ptr &tmp : std::as_const(d->devices)) {
         if (tmp->id() != device->id())
             names << tmp->displayName();
     }
 
-    device->setDisplayName(Utils::makeUniquelyNumbered(device->displayName(), names));
+    // TODO: make it thread safe?
+    device->settings()->displayName.setValue(
+        Utils::makeUniquelyNumbered(device->displayName(), names));
 
     const int pos = d->indexForId(device->id());
 
@@ -252,27 +269,42 @@ void DeviceManager::addDevice(const IDevice::ConstPtr &_device)
         d->clonedInstance->addDevice(device->clone());
 
     if (pos >= 0) {
-        d->devices[pos] = device;
+        {
+            QMutexLocker locker(&d->mutex);
+            d->devices[pos] = device;
+        }
         emit deviceUpdated(device->id());
     } else {
-        d->devices << device;
+        {
+            QMutexLocker locker(&d->mutex);
+            d->devices << device;
+        }
         emit deviceAdded(device->id());
+
+        if (FSEngine::isAvailable())
+            FSEngine::addDevice(device->rootPath());
     }
 
     emit updated();
 }
 
-void DeviceManager::removeDevice(Core::Id id)
+void DeviceManager::removeDevice(Id id)
 {
     const IDevice::Ptr device = mutableDevice(id);
     QTC_ASSERT(device, return);
     QTC_ASSERT(this != instance() || device->isAutoDetected(), return);
 
     const bool wasDefault = d->defaultDevices.value(device->type()) == device->id();
-    const Core::Id deviceType = device->type();
-    d->devices.removeAt(d->indexForId(id));
+    const Id deviceType = device->type();
+    {
+        QMutexLocker locker(&d->mutex);
+        d->devices.removeAt(d->indexForId(id));
+    }
     emit deviceRemoved(device->id());
 
+    if (FSEngine::isAvailable())
+        FSEngine::removeDevice(device->rootPath());
+
     if (wasDefault) {
         for (int i = 0; i < d->devices.count(); ++i) {
             if (deviceAt(i)->type() == deviceType) {
@@ -288,7 +320,7 @@ void DeviceManager::removeDevice(Core::Id id)
     emit updated();
 }
 
-void DeviceManager::setDeviceState(Core::Id deviceId, IDevice::DeviceState deviceState)
+void DeviceManager::setDeviceState(Id deviceId, IDevice::DeviceState deviceState)
 {
     // To see the state change in the DeviceSettingsWidget. This has to happen before
     // the pos check below, in case the device is only present in the cloned instance.
@@ -302,6 +334,7 @@ void DeviceManager::setDeviceState(Core::Id deviceId, IDevice::DeviceState devic
     if (device->deviceState() == deviceState)
         return;
 
+    // TODO: make it thread safe?
     device->setDeviceState(deviceState);
     emit deviceUpdated(deviceId);
     emit updated();
@@ -312,7 +345,33 @@ bool DeviceManager::isLoaded() const
     return d->writer;
 }
 
-void DeviceManager::setDefaultDevice(Core::Id id)
+// Thread safe
+IDevice::ConstPtr DeviceManager::deviceForPath(const FilePath &path)
+{
+    const QList<IDevice::Ptr> devices = instance()->d->deviceList();
+
+    if (path.scheme() == u"device") {
+        for (const IDevice::Ptr &dev : devices) {
+            if (path.host() == dev->id().toString())
+                return dev;
+        }
+        return {};
+    }
+
+    for (const IDevice::Ptr &dev : devices) {
+        // TODO: ensure handlesFile is thread safe
+        if (dev->handlesFile(path))
+            return dev;
+    }
+    return {};
+}
+
+IDevice::ConstPtr DeviceManager::defaultDesktopDevice()
+{
+    return m_instance->defaultDevice(Constants::DESKTOP_DEVICE_TYPE);
+}
+
+void DeviceManager::setDefaultDevice(Id id)
 {
     QTC_ASSERT(this != instance(), return);
 
@@ -328,28 +387,93 @@ void DeviceManager::setDefaultDevice(Core::Id id)
     emit updated();
 }
 
-const IDeviceFactory *DeviceManager::restoreFactory(const QVariantMap &map)
-{
-    const Core::Id deviceType = IDevice::typeFromMap(map);
-    IDeviceFactory *factory = Utils::findOrDefault(IDeviceFactory::allDeviceFactories(),
-        [&map, deviceType](IDeviceFactory *factory) {
-            return factory->canRestore(map) && factory->deviceType() == deviceType;
-        });
-
-    if (!factory)
-        qWarning("Warning: No factory found for device '%s' of type '%s'.",
-                 qPrintable(IDevice::idFromMap(map).toString()),
-                 qPrintable(IDevice::typeFromMap(map).toString()));
-    return factory;
-}
-
 DeviceManager::DeviceManager(bool isInstance) : d(std::make_unique<DeviceManagerPrivate>())
 {
-    if (isInstance) {
-        QTC_ASSERT(!m_instance, return);
-        m_instance = this;
-        connect(Core::ICore::instance(), &Core::ICore::saveSettingsRequested, this, &DeviceManager::save);
-    }
+    QTC_ASSERT(isInstance == !m_instance, return);
+
+    if (!isInstance)
+        return;
+
+    m_instance = this;
+    connect(Core::ICore::instance(), &Core::ICore::saveSettingsRequested,
+            this, &DeviceManager::save);
+
+    DeviceFileHooks &deviceHooks = DeviceFileHooks::instance();
+
+    deviceHooks.isSameDevice = [](const FilePath &left, const FilePath &right) {
+        auto leftDevice = DeviceManager::deviceForPath(left);
+        auto rightDevice = DeviceManager::deviceForPath(right);
+
+        return leftDevice == rightDevice;
+    };
+
+    deviceHooks.localSource = [](const FilePath &file) -> expected_str<FilePath> {
+        auto device = DeviceManager::deviceForPath(file);
+        if (!device)
+            return make_unexpected(Tr::tr("No device for path \"%1\"").arg(file.toUserOutput()));
+        return device->localSource(file);
+    };
+
+    deviceHooks.fileAccess = [](const FilePath &filePath) -> expected_str<DeviceFileAccess *> {
+        if (!filePath.needsDevice())
+            return DesktopDeviceFileAccess::instance();
+        IDevice::ConstPtr device = DeviceManager::deviceForPath(filePath);
+        if (!device) {
+            return make_unexpected(
+                Tr::tr("No device found for path \"%1\"").arg(filePath.toUserOutput()));
+        }
+        DeviceFileAccess *fileAccess = device->fileAccess();
+        if (!fileAccess) {
+            return make_unexpected(
+                Tr::tr("No file access for device \"%1\"").arg(device->displayName()));
+        }
+        return fileAccess;
+    };
+
+    deviceHooks.environment = [](const FilePath &filePath) -> expected_str<Environment> {
+        auto device = DeviceManager::deviceForPath(filePath);
+        if (!device) {
+            return make_unexpected(
+                Tr::tr("No device found for path \"%1\"").arg(filePath.toUserOutput()));
+        }
+        return device->systemEnvironmentWithError();
+    };
+
+    deviceHooks.deviceDisplayName = [](const FilePath &filePath) {
+        auto device = DeviceManager::deviceForPath(filePath);
+        if (device)
+            return device->displayName();
+        return filePath.host().toString();
+    };
+
+    deviceHooks.ensureReachable = [](const FilePath &filePath, const FilePath &other) {
+        auto device = DeviceManager::deviceForPath(filePath);
+        QTC_ASSERT(device, return false);
+        return device->ensureReachable(other);
+    };
+
+    deviceHooks.openTerminal = [](const FilePath &filePath, const Environment &env) {
+        auto device = DeviceManager::deviceForPath(filePath);
+        QTC_ASSERT(device, return);
+        device->openTerminal(env, filePath);
+    };
+
+    deviceHooks.osType = [](const FilePath &filePath) {
+        auto device = DeviceManager::deviceForPath(filePath);
+        if (!device)
+            return OsTypeLinux;
+        return device->osType();
+    };
+
+    DeviceProcessHooks processHooks;
+
+    processHooks.processImplHook = [](const FilePath &filePath) -> ProcessInterface * {
+        auto device = DeviceManager::deviceForPath(filePath);
+        QTC_ASSERT(device, return nullptr);
+        return device->createProcessInterface();
+    };
+
+    Process::setRemoteProcessHooks(processHooks);
 }
 
 DeviceManager::~DeviceManager()
@@ -366,7 +490,15 @@ IDevice::ConstPtr DeviceManager::deviceAt(int idx) const
     return d->devices.at(idx);
 }
 
-IDevice::Ptr DeviceManager::mutableDevice(Core::Id id) const
+void DeviceManager::forEachDevice(const std::function<void(const IDeviceConstPtr &)> &func) const
+{
+    const QList<IDevice::Ptr> devices = d->deviceList();
+
+    for (const IDevice::Ptr &device : devices)
+        func(device);
+}
+
+IDevice::Ptr DeviceManager::mutableDevice(Id id) const
 {
     const int index = d->indexForId(id);
     return index == -1 ? IDevice::Ptr() : d->devices.at(index);
@@ -379,15 +511,15 @@ bool DeviceManager::hasDevice(const QString &name) const
     });
 }
 
-IDevice::ConstPtr DeviceManager::find(Core::Id id) const
+IDevice::ConstPtr DeviceManager::find(Id id) const
 {
     const int index = d->indexForId(id);
     return index == -1 ? IDevice::ConstPtr() : deviceAt(index);
 }
 
-IDevice::ConstPtr DeviceManager::defaultDevice(Core::Id deviceType) const
+IDevice::ConstPtr DeviceManager::defaultDevice(Id deviceType) const
 {
-    const Core::Id id = d->defaultDevices.value(deviceType);
+    const Id id = d->defaultDevices.value(deviceType);
     return id.isValid() ? find(id) : IDevice::ConstPtr();
 }
 
@@ -395,7 +527,7 @@ IDevice::ConstPtr DeviceManager::defaultDevice(Core::Id deviceType) const
 
 
 #ifdef WITH_TESTS
-#include <projectexplorer/projectexplorer.h>
+#include <projectexplorer/projectexplorer_test.h>
 #include <QSignalSpy>
 #include <QTest>
 #include <QUuid>
@@ -407,23 +539,20 @@ class TestDevice : public IDevice
 public:
     TestDevice()
     {
-        setupId(AutoDetected, Core::Id::fromString(QUuid::createUuid().toString()));
+        setupId(AutoDetected, Id::fromString(QUuid::createUuid().toString()));
         setType(testTypeId());
         setMachineType(Hardware);
+        setOsType(HostOsInfo::hostOs());
+        setDisplayType("blubb");
     }
 
-    static Core::Id testTypeId() { return "TestType"; }
+    static Id testTypeId() { return "TestType"; }
+
 private:
-    QString displayType() const override { return QLatin1String("blubb"); }
     IDeviceWidget *createWidget() override { return nullptr; }
-    DeviceProcessSignalOperation::Ptr signalOperation() const override
-    {
-        return DeviceProcessSignalOperation::Ptr();
-    }
-    Utils::OsType osType() const override { return Utils::HostOsInfo::hostOs(); }
 };
 
-class TestDeviceFactory : public IDeviceFactory
+class TestDeviceFactory final : public IDeviceFactory
 {
 public:
     TestDeviceFactory() : IDeviceFactory(TestDevice::testTypeId())
@@ -432,12 +561,12 @@ public:
     }
 };
 
-void ProjectExplorerPlugin::testDeviceManager()
+void ProjectExplorerTest::testDeviceManager()
 {
     TestDeviceFactory factory;
 
     TestDevice::Ptr dev = IDevice::Ptr(new TestDevice);
-    dev->setDisplayName(QLatin1String("blubbdiblubbfurz!"));
+    dev->settings()->displayName.setValue(QLatin1String("blubbdiblubbfurz!"));
     QVERIFY(dev->isAutoDetected());
     QCOMPARE(dev->deviceState(), IDevice::DeviceStateUnknown);
     QCOMPARE(dev->type(), TestDevice::testTypeId());
@@ -498,7 +627,7 @@ void ProjectExplorerPlugin::testDeviceManager()
     TestDevice::Ptr dev3 = IDevice::Ptr(new TestDevice);
     QVERIFY(dev->id() != dev3->id());
 
-    dev3->setDisplayName(dev->displayName());
+    dev3->settings()->displayName.setValue(dev->displayName());
     mgr->addDevice(dev3);
     QCOMPARE(mgr->deviceAt(mgr->deviceCount() - 1)->displayName(),
              QString(dev3->displayName() + QLatin1Char('2')));
