@@ -1,34 +1,10 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of Qt Creator.
-**
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 as published by the Free Software
-** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-****************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
 
 #include "qbsprofilemanager.h"
 
 #include "defaultpropertyprovider.h"
-#include "qbsproject.h"
-#include "qbsprojectmanagerconstants.h"
-#include "qbsprojectmanagerplugin.h"
+#include "qbsprojectmanagertr.h"
 #include "qbssettings.h"
 
 #include <coreplugin/icore.h>
@@ -38,9 +14,9 @@
 #include <projectexplorer/projectexplorer.h>
 #include <qmljstools/qmljstoolsconstants.h>
 #include <qtsupport/baseqtversion.h>
-#include <qtsupport/qtkitinformation.h>
-#include <utils/qtcassert.h>
+#include <qtsupport/qtkitaspect.h>
 #include <utils/qtcprocess.h>
+#include <utils/qtcassert.h>
 
 #include <QCryptographicHash>
 #include <QJSEngine>
@@ -81,7 +57,7 @@ QString toJSLiteral(const QVariant &val)
 {
     if (!val.isValid())
         return QString("undefined");
-    if (val.type() == QVariant::List || val.type() == QVariant::StringList) {
+    if (val.typeId() == QMetaType::QVariantList || val.typeId() == QMetaType::QStringList) {
         QString res;
         const auto list = val.toList();
         for (const QVariant &child : list) {
@@ -92,7 +68,7 @@ QString toJSLiteral(const QVariant &val)
         res.append(']');
         return res;
     }
-    if (val.type() == QVariant::Map) {
+    if (val.typeId() == QMetaType::QVariantMap) {
         const QVariantMap &vm = val.toMap();
         QString str("{");
         for (auto it = vm.begin(); it != vm.end(); ++it) {
@@ -103,28 +79,34 @@ QString toJSLiteral(const QVariant &val)
         str += '}';
         return str;
     }
-    if (val.type() == QVariant::Bool)
+    if (val.typeId() == QMetaType::Bool)
         return toJSLiteral(val.toBool());
-    if (val.canConvert(QVariant::String))
+    if (val.canConvert(QMetaType(QMetaType::QString)))
         return toJSLiteral(val.toString());
     return QString::fromLatin1("Unconvertible type %1").arg(QLatin1String(val.typeName()));
 }
 
-
-static QbsProfileManager *m_instance = nullptr;
+static PropertyProvider &defaultPropertyProvider()
+{
+    static DefaultPropertyProvider theDefaultPropertyProvider;
+    return theDefaultPropertyProvider;
+}
 
 static QString kitNameKeyInQbsSettings(const ProjectExplorer::Kit *kit)
 {
     return "preferences.qtcreator.kit." + kit->id().toString();
 }
 
-QbsProfileManager::QbsProfileManager() : m_defaultPropertyProvider(new DefaultPropertyProvider)
+QbsProfileManager::QbsProfileManager()
 {
-    m_instance = this;
-
     setObjectName(QLatin1String("QbsProjectManager"));
-    connect(ProjectExplorer::KitManager::instance(), &ProjectExplorer::KitManager::kitsLoaded, this,
-            [this]() { m_kitsToBeSetupForQbs = ProjectExplorer::KitManager::kits(); } );
+
+    if (ProjectExplorer::KitManager::instance()->isLoaded()) {
+        m_kitsToBeSetupForQbs = ProjectExplorer::KitManager::kits();
+    } else {
+        connect(ProjectExplorer::KitManager::instance(), &ProjectExplorer::KitManager::kitsLoaded,
+                this, [this] { m_kitsToBeSetupForQbs = ProjectExplorer::KitManager::kits(); } );
+    }
     connect(ProjectExplorer::KitManager::instance(), &ProjectExplorer::KitManager::kitAdded, this,
             &QbsProfileManager::addProfileFromKit);
     connect(ProjectExplorer::KitManager::instance(), &ProjectExplorer::KitManager::kitUpdated, this,
@@ -135,15 +117,12 @@ QbsProfileManager::QbsProfileManager() : m_defaultPropertyProvider(new DefaultPr
             this, &QbsProfileManager::updateAllProfiles);
 }
 
-QbsProfileManager::~QbsProfileManager()
-{
-    delete m_defaultPropertyProvider;
-    m_instance = nullptr;
-}
+QbsProfileManager::~QbsProfileManager() = default;
 
 QbsProfileManager *QbsProfileManager::instance()
 {
-    return m_instance;
+    static QbsProfileManager theQbsProfileManager;
+    return &theQbsProfileManager;
 }
 
 QString QbsProfileManager::ensureProfileForKit(const ProjectExplorer::Kit *k)
@@ -158,8 +137,8 @@ void QbsProfileManager::updateProfileIfNecessary(const ProjectExplorer::Kit *kit
 {
     // kit in list <=> profile update is necessary
     // Note that the const_cast is safe, as we do not call any non-const methods on the object.
-    if (m_instance->m_kitsToBeSetupForQbs.removeOne(const_cast<ProjectExplorer::Kit *>(kit)))
-        m_instance->addProfileFromKit(kit);
+    if (instance()->m_kitsToBeSetupForQbs.removeOne(const_cast<ProjectExplorer::Kit *>(kit)))
+        instance()->addProfileFromKit(kit);
 }
 
 void QbsProfileManager::updateAllProfiles()
@@ -175,8 +154,8 @@ void QbsProfileManager::addProfileFromKit(const ProjectExplorer::Kit *k)
     runQbsConfig(QbsConfigOp::Set, kitNameKeyInQbsSettings(k), name);
 
     // set up properties:
-    QVariantMap data = m_defaultPropertyProvider->properties(k, QVariantMap());
-    for (PropertyProvider *provider : qAsConst(g_propertyProviders)) {
+    QVariantMap data = defaultPropertyProvider().properties(k, QVariantMap());
+    for (PropertyProvider *provider : std::as_const(g_propertyProviders)) {
         if (provider->canHandle(k))
             data = provider->properties(k, data);
     }
@@ -195,8 +174,8 @@ void QbsProfileManager::addProfileFromKit(const ProjectExplorer::Kit *k)
 
 void QbsProfileManager::handleKitUpdate(ProjectExplorer::Kit *kit)
 {
-    m_kitsToBeSetupForQbs.removeOne(kit);
-    addProfileFromKit(kit);
+    if (!m_kitsToBeSetupForQbs.contains(kit))
+        addProfileFromKit(kit);
 }
 
 void QbsProfileManager::handleKitRemoval(ProjectExplorer::Kit *kit)
@@ -244,18 +223,19 @@ QString QbsProfileManager::runQbsConfig(QbsConfigOp op, const QString &key, cons
     const Utils::FilePath qbsConfigExe = QbsSettings::qbsConfigFilePath();
     if (qbsConfigExe.isEmpty() || !qbsConfigExe.exists())
         return {};
-    Utils::QtcProcess qbsConfig;
+    Utils::Process qbsConfig;
     qbsConfig.setCommand({qbsConfigExe, args});
     qbsConfig.start();
-    if (!qbsConfig.waitForStarted(3000) || !qbsConfig.waitForFinished(5000)) {
+    using namespace std::chrono_literals;
+    if (!qbsConfig.waitForFinished(5s)) {
         Core::MessageManager::writeFlashing(
-            tr("Failed to run qbs config: %1").arg(qbsConfig.errorString()));
+            Tr::tr("Failed to run qbs config: %1").arg(qbsConfig.errorString()));
     } else if (qbsConfig.exitCode() != 0) {
         Core::MessageManager::writeFlashing(
-            tr("Failed to run qbs config: %1")
-                .arg(QString::fromLocal8Bit(qbsConfig.readAllStandardError())));
+            Tr::tr("Failed to run qbs config: %1")
+                .arg(QString::fromLocal8Bit(qbsConfig.rawStdErr())));
     }
-    return QString::fromLocal8Bit(qbsConfig.readAllStandardOutput()).trimmed();
+    return QString::fromLocal8Bit(qbsConfig.rawStdOut()).trimmed();
 }
 
 QVariant fromJSLiteral(const QString &str)
