@@ -1,27 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2020 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of Qt Creator.
-**
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 as published by the Free Software
-** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-****************************************************************************/
+// Copyright (C) 2020 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
 #include "stylesheetmerger.h"
 
 #include <abstractview.h>
@@ -34,11 +12,19 @@
 #include <nodelistproperty.h>
 #include <nodemetainfo.h>
 #include <nodeproperty.h>
+#include <plaintexteditmodifier.h>
+#include <rewriterview.h>
 #include <variantproperty.h>
 
+#include <utils/fileutils.h>
+#include <utils/qtcassert.h>
+
+#include <QPlainTextEdit>
 #include <QQueue>
 #include <QRegularExpression>
 
+#include <memory>
+
 namespace {
 
 QPoint pointForModelNode(const QmlDesigner::ModelNode &node)
@@ -81,7 +67,7 @@ static void splitIdInBaseNameAndNumber(const QString &id, QString *baseId, int *
 {
 
     int counter = 0;
-    while (counter < id.count()) {
+    while (counter < id.size()) {
         bool canConvertToInteger = false;
         int newNumber = id.right(counter + 1).toInt(&canConvertToInteger);
         if (canConvertToInteger)
@@ -92,7 +78,7 @@ static void splitIdInBaseNameAndNumber(const QString &id, QString *baseId, int *
         counter++;
     }
 
-    *baseId = id.left(id.count() - counter);
+    *baseId = id.left(id.size() - counter);
 }
 
 void StylesheetMerger::syncNodeProperties(ModelNode &outputNode, const ModelNode &inputNode, bool skipDuplicates)
@@ -131,9 +117,8 @@ void StylesheetMerger::syncVariantProperties(ModelNode &outputNode, const ModelN
 
 void StylesheetMerger::syncAuxiliaryProperties(ModelNode &outputNode, const ModelNode &inputNode)
 {
-    auto tmp = inputNode.auxiliaryData();
-    for (auto iter = tmp.begin(); iter != tmp.end(); ++iter)
-        outputNode.setAuxiliaryData(iter.key(), iter.value());
+    for (const auto &[key, value] : inputNode.auxiliaryData())
+        outputNode.setAuxiliaryData(AuxiliaryDataKeyView{key}, value);
 }
 
 void StylesheetMerger::syncBindingProperties(ModelNode &outputNode, const ModelNode &inputNode)
@@ -155,14 +140,16 @@ void StylesheetMerger::syncId(ModelNode &outputNode, ModelNode &inputNode)
 
 void StylesheetMerger::setupIdRenamingHash()
 {
-    for (const ModelNode &node : m_templateView->rootModelNode().allSubModelNodesAndThisNode()) {
+    const QList<ModelNode> &nodes = m_templateView->rootModelNode().allSubModelNodesAndThisNode();
+    for (const ModelNode &node : nodes) {
         if (!node.id().isEmpty()) {
             QString newId = node.id();
             QString baseId;
             int number = 1;
             splitIdInBaseNameAndNumber(newId, &baseId, &number);
 
-            while (m_templateView->hasId(newId) || m_idReplacementHash.values().contains(newId)) {
+            while (m_templateView->hasId(newId) || std::find(m_idReplacementHash.cbegin(),
+                        m_idReplacementHash.cend(), newId) != m_idReplacementHash.cend()) {
                 newId = "stylesheet_auto_merge_" + baseId + QString::number(number);
                 number++;
             }
@@ -175,7 +162,6 @@ void StylesheetMerger::setupIdRenamingHash()
 ModelNode StylesheetMerger::createReplacementNode(const ModelNode& styleNode, ModelNode &modelNode)
 {
     QList<QPair<PropertyName, QVariant> > propertyList;
-    QList<QPair<PropertyName, QVariant> > auxPropertyList;
     NodeMetaInfo nodeMetaInfo = m_templateView->model()->metaInfo(styleNode.type());
 
     for (const VariantProperty &variantProperty : modelNode.variantProperties()) {
@@ -185,14 +171,26 @@ ModelNode StylesheetMerger::createReplacementNode(const ModelNode& styleNode, Mo
             continue;
         propertyList.append(QPair<PropertyName, QVariant>(variantProperty.name(), variantProperty.value()));
     }
-    ModelNode newNode(m_templateView->createModelNode(styleNode.type(), nodeMetaInfo.majorVersion(), nodeMetaInfo.minorVersion(),
-                                                      propertyList, auxPropertyList, styleNode.nodeSource(), styleNode.nodeSourceType()));
+
+#ifdef QDS_USE_PROJECTSTORAGE
+    ModelNode newNode(m_templateView->createModelNode(
+        styleNode.type(), propertyList, {}, styleNode.nodeSource(), styleNode.nodeSourceType()));
+#else
+    ModelNode newNode(m_templateView->createModelNode(styleNode.type(),
+                                                      nodeMetaInfo.majorVersion(),
+                                                      nodeMetaInfo.minorVersion(),
+                                                      propertyList,
+                                                      {},
+                                                      styleNode.nodeSource(),
+                                                      styleNode.nodeSourceType()));
+#endif
 
     syncAuxiliaryProperties(newNode, modelNode);
     syncBindingProperties(newNode, modelNode);
     syncId(newNode, modelNode);
     syncNodeProperties(newNode, modelNode);
     syncNodeListProperties(newNode, modelNode);
+    mergeStates(newNode, modelNode);
 
     return newNode;
 }
@@ -413,6 +411,84 @@ void StylesheetMerger::parseTemplateOptions()
     }
 }
 
+void StylesheetMerger::syncStateNode(ModelNode &outputState, const ModelNode &inputState) const
+{
+    auto addProperty = [](ModelNode &n, const AbstractProperty &p) {
+        if (n.hasProperty(p.name()))
+            return; // Do not ovewrite. Only merge when property not defined in template.
+        if (p.isBindingProperty())
+            n.bindingProperty(p.name()).setExpression(p.toBindingProperty().expression());
+        else
+            n.variantProperty(p.name()).setValue(p.toVariantProperty().value());
+    };
+
+    addProperty(outputState, inputState.property("when"));
+    addProperty(outputState, inputState.property("extend"));
+
+    auto changeSetKey = [](const ModelNode &n) {
+        return QString("%1::%2").arg(QString::fromUtf8(n.type()),
+                                     n.bindingProperty("target").expression());
+    };
+
+    // Collect change sets already defined in the output state.
+    std::map<QString, ModelNode> outputChangeSets;
+    for (ModelNode propChange : outputState.directSubModelNodes())
+        outputChangeSets.insert({changeSetKey(propChange), propChange});
+
+    // Merge the child nodes of the states i.e. AnchorChanges, PropertyChanges, etc.
+    for (ModelNode inputChangeset : inputState.directSubModelNodes()) {
+        const QString key = changeSetKey(inputChangeset);
+        const auto itr = outputChangeSets.find(key);
+        ModelNode changeSet;
+        if (itr != outputChangeSets.end()) {
+            changeSet = itr->second;
+        } else {
+            const QByteArray typeName = inputChangeset.type();
+#ifdef QDS_USE_PROJECTSTORAGE
+            changeSet = m_templateView->createModelNode(typeName);
+#else
+            NodeMetaInfo metaInfo = m_templateView->model()->metaInfo(typeName);
+            int major = metaInfo.majorVersion();
+            int minor = metaInfo.minorVersion();
+            changeSet = m_templateView->createModelNode(typeName, major, minor);
+#endif
+            outputState.nodeListProperty("changes").reparentHere(changeSet);
+            outputChangeSets.insert({key, changeSet});
+        }
+
+        for (const auto &p : inputChangeset.properties())
+            addProperty(changeSet, p);
+    }
+}
+
+void StylesheetMerger::mergeStates(ModelNode &outputNode, const ModelNode &inputNode) const
+{
+    QMap<QString, ModelNode> outputStates;
+    for (auto stateNode : outputNode.nodeListProperty("states").toModelNodeList()) {
+        const QString name = stateNode.variantProperty("name").value().toString();
+        if (name.isEmpty())
+            continue;
+        outputStates[name] = stateNode;
+    }
+
+    for (auto inputStateNode : inputNode.nodeListProperty("states").toModelNodeList()) {
+        const QString name = inputStateNode.variantProperty("name").value().toString();
+        try {
+            if (outputStates.contains(name)) {
+                syncStateNode(outputStates[name], inputStateNode);
+                continue;
+            }
+            ModelMerger merger(m_templateView);
+            ModelNode stateClone = merger.insertModel(inputStateNode);
+            if (stateClone.isValid())
+                outputNode.nodeListProperty("states").reparentHere(stateClone);
+        } catch (Exception &e) {
+            qDebug().noquote() << "Exception while merging states.";
+            e.createWarning();
+        }
+    }
+}
+
 void StylesheetMerger::merge()
 {
     ModelNode templateRootNode = m_templateView->rootModelNode();
@@ -433,6 +509,8 @@ void StylesheetMerger::merge()
         return;
     }
 
+    mergeStates(templateRootNode, styleRootNode);
+
     QQueue<ModelNode> replacementNodes;
 
     QList<ModelNode> directRootSubNodes = styleRootNode.directSubModelNodes();
@@ -525,4 +603,88 @@ void StylesheetMerger::merge()
         }
     }
 }
+
+void StylesheetMerger::styleMerge(const Utils::FilePath &templateFile,
+                                  Model *model,
+                                  ExternalDependenciesInterface &externalDependencies)
+{
+    Utils::FileReader reader;
+
+    QTC_ASSERT(reader.fetch(templateFile), return );
+    const QString qmlTemplateString = QString::fromUtf8(reader.data());
+    StylesheetMerger::styleMerge(qmlTemplateString, model, externalDependencies);
 }
+
+void StylesheetMerger::styleMerge(const QString &qmlTemplateString,
+                                  Model *model,
+                                  ExternalDependenciesInterface &externalDependencies)
+{
+    Model *parentModel = model;
+
+    QTC_ASSERT(parentModel, return );
+
+#ifdef QDS_USE_PROJECTSTORAGE
+    auto templateModel = model->createModel("Item");
+#else
+    auto templateModel(Model::create("QtQuick.Item", 2, 1, parentModel));
+#endif
+    Q_ASSERT(templateModel.get());
+
+    templateModel->setFileUrl(parentModel->fileUrl());
+
+    QPlainTextEdit textEditTemplate;
+    QString imports;
+
+    for (const Import &import : parentModel->imports()) {
+        imports += QStringLiteral("import ") + import.toString(true) + QLatin1Char(';')
+                   + QLatin1Char('\n');
+    }
+
+    textEditTemplate.setPlainText(imports + qmlTemplateString);
+    NotIndentingTextEditModifier textModifierTemplate(&textEditTemplate);
+
+    std::unique_ptr<RewriterView> templateRewriterView = std::make_unique<RewriterView>(
+        externalDependencies, RewriterView::Amend);
+    templateRewriterView->setTextModifier(&textModifierTemplate);
+    templateModel->attachView(templateRewriterView.get());
+    templateRewriterView->setCheckSemanticErrors(false);
+    templateRewriterView->setPossibleImportsEnabled(false);
+
+    ModelNode templateRootNode = templateRewriterView->rootModelNode();
+    QTC_ASSERT(templateRootNode.isValid(), return );
+
+#ifdef QDS_USE_PROJECTSTORAGE
+    auto styleModel = model->createModel("Item");
+#else
+    auto styleModel(Model::create("QtQuick.Item", 2, 1, parentModel));
+    Q_ASSERT(styleModel.get());
+#endif
+    styleModel->setFileUrl(parentModel->fileUrl());
+
+    QPlainTextEdit textEditStyle;
+    RewriterView *parentRewriterView = parentModel->rewriterView();
+    QTC_ASSERT(parentRewriterView, return );
+    textEditStyle.setPlainText(parentRewriterView->textModifierContent());
+    NotIndentingTextEditModifier textModifierStyle(&textEditStyle);
+
+    std::unique_ptr<RewriterView> styleRewriterView = std::make_unique<RewriterView>(
+        externalDependencies, RewriterView::Amend);
+    styleRewriterView->setTextModifier(&textModifierStyle);
+    styleModel->attachView(styleRewriterView.get());
+
+    StylesheetMerger merger(templateRewriterView.get(), styleRewriterView.get());
+
+    try {
+        merger.merge();
+    } catch (Exception &e) {
+        e.showException();
+    }
+
+    try {
+        parentRewriterView->textModifier()->textDocument()->setPlainText(
+            templateRewriterView->textModifierContent());
+    } catch (Exception &e) {
+        e.showException();
+    }
+}
+} // namespace QmlDesigner
